#include "Vector.h"
#include "CSRMatrix.h"

//!!! Если параметров много, то они форматируются в столбик
CSRMatrix::CSRMatrix()
    : elements(nullptr)
    , columns(nullptr)
    , rowPointers(nullptr)
    , elementsSize(0)
    , columnsSize(0)
    , rowPointersSize(0)
{}

// 
CSRMatrix::CSRMatrix(
    double *element,
    int    *column,
    int    *rowPointer, 
    size_t  elementsSize,
    size_t  columnsSize, 
    size_t  rowPointersSize)
    : elementsSize(elementsSize)
    , columnsSize(columnsSize)
    , rowPointersSize(rowPointersSize)
{
    elements = new double[elementsSize];
    columns = new int[columnsSize];
    rowPointers = new int[rowPointersSize];
    for (int i = 0; i < elementsSize; ++i)
    {
        this->elements[i] = element[i];
    }
    for (int i = 0; i < columnsSize; ++i)
    {
        this->columns[i] = column[i];
    }
    for (int i = 0; i < rowPointersSize; ++i)
    {
        this->rowPointers[i] = rowPointer[i];
    }
    std::cout << "\n" << "  " << this << "\n";
}

//  
CSRMatrix::CSRMatrix(const CSRMatrix& other) : elementsSize(other.elementsSize), columnsSize(other.columnsSize), rowPointersSize(other.rowPointersSize)
{
    elements = new double[elementsSize];
    columns = new int[columnsSize];
    rowPointers = new int[rowPointersSize];
    for (int i = 0; i < elementsSize; ++i)
    {
        this->elements[i] = other.elements[i];
    }
    for (int i = 0; i < columnsSize; ++i)
    {
        this->columns[i] = other.columns[i];
    }
    for (int i = 0; i < rowPointersSize; ++i)
    {
        this->rowPointers[i] = other.rowPointers[i];
    }
    std::cout << "\n" << "  ! " << this << "\n";
}

// 
CSRMatrix::~CSRMatrix() 
{
    delete[] elements;
    delete[] columns;
    delete[] rowPointers;
    std::cout << "\n" << " ! " << this << "\n";
}

// ,     
size_t CSRMatrix::GetRows() const 
{
    return rowPointersSize;
}

// ,     
size_t CSRMatrix::GetColumns() const 
{
    return columnsSize;
}

// ,    
size_t CSRMatrix::NonZeroElements() const
{
    return elementsSize;
}

//   =  
CSRMatrix& CSRMatrix::operator = (const CSRMatrix& other)
{
    if (this != &other)
    {
        //!!! Выделять память заново нужно только, если размеры не совпадают
        
        
        if (this->elements != nullptr) //!!! Лишняя проврека
        {
            delete[] this->elements;
        }
        this->elementsSize = other.elementsSize;
        this->elements = new double[other.elementsSize];
        for (int i = 0; i < elementsSize; ++i)
        {
            this->elements[i] = other.elements[i];
        }
        
        if (this->columns != nullptr)
        {
            delete[] this->columns;
        }
        this->columnsSize = other.columnsSize;
        this->columns = new int[other.columnsSize];
        for (int i = 0; i < columnsSize; ++i)
        {
            this->columns[i] = other.columns[i];
        }

        if (this->rowPointers != nullptr)
        {
            delete[] this->rowPointers;
        }
        this->rowPointersSize = other.rowPointersSize;
        this->rowPointers = new int[other.rowPointersSize];
        for (int i = 0; i < rowPointersSize; ++i)
        {
            this->rowPointers[i] = other.rowPointers[i];
        }
    }
    std::cout << "\n" << "   = " << this << "\n"; //!!! Убрать
    return *this;
}

//    (  )  
CSRMatrix CSRMatrix::operator*(double value) const 
{
    //!!! Почему изменяется первый операнд
    
    for (int i = 0; i < elementsSize; ++i) 
    {
        this->elements[i] *= value;
    }
    return *this;
}

//       (  )
CSRMatrix operator *(double value, const CSRMatrix& matrix)
{
    CSRMatrix temp(matrix);
    for (int i = 0; i < temp.elementsSize; ++i)
    {
        temp.elements[i] *= value;
    }
    return temp;
}

//      
Vector CSRMatrix::operator*(const Vector& vector) const 
{
    int size = vector.getsize();
    if (rowPointersSize - 1 != size) 
    {
        throw InCompatibleDimException();
    }
    Vector temp(size);
    for (int row = 0; row < size; ++row) {
        for (int i = rowPointers[row]; i < rowPointers[row + 1]; ++i) 
        {
            temp[row] += elements[i] * elements[columns[i]];
        }
    }
    return temp;

}

//      
Vector operator*(const Vector& vector, const CSRMatrix& matrix)
{
    int matrixSize = matrix.GetRows()-1;
    int vectorSize = vector.getsize();
    Vector temp(matrixSize);
    if (matrixSize != vectorSize) 
    {
        throw InCompatibleDimException();
    }
    for (int row = 0; row < matrixSize; ++row) 
    {
        for (int i = matrix.rowPointers[row]; i < matrix.rowPointers[row + 1]; ++i) 
        {
            temp[matrix.columns[i]] += matrix.elements[i] * vector[row]; //!!! Сомнительно
        }
    }
    return temp;
} 

//      
std::istream& operator >> (std::istream& in, CSRMatrix& matrix)
{
    //!!! Оператор должен спросить у пользователя новый размер и,
    //!!! если он отличен от страрого, очистить память и выделить новую

    
    std::cout << "   " << "\n";
    for (int i = 0; i <matrix.elementsSize; ++i)
    {
        in >> matrix.elements[i];
    }
    std::cout << " ,      " << "\n";
    for (int i = 0; i < matrix.columnsSize; ++i)
    {
        in >> matrix.columns[i];
    }
    std::cout << "   " << "\n";
    for (int i = 0; i < matrix.rowPointersSize; ++i)
    {
        in >> matrix.rowPointers[i];
    }
    return in;
}

//      
std::ostream& operator <<(std::ostream& out, const CSRMatrix& matrix)
{
    //!!! НА ЗАЩИТУ:
    //!!! Реализовать данный оператор для вывода полной матрицы (со всеми нулями),
    //!!! но БЕЗ выделения памяти под всю матрицу
    
    for (int i = 0; i < matrix.elementsSize; ++i)
    {
        out << matrix.elements[i] << " ";
    }
    std::cout << "   \n"; //!!! Что тут делает std::cout ?
    
    for (int i = 0; i < matrix.columnsSize; ++i) 
    {
        out << matrix.columns[i] << " ";
    }
    std::cout << "  ,      \n";//!!! Что тут делает std::cout ?
    
    for (int i = 0; i < matrix.rowPointersSize; ++i) 
    {
        out << matrix.rowPointers[i] << " ";
    }
    std::cout << "   \n";//!!! Что тут делает std::cout ?
    return out;
}





